package com.loafle.overflow.proxy.db;

import com.loafle.overflow.db.api.DBGrpc;
import com.loafle.overflow.db.api.DBInput;
import com.loafle.overflow.db.api.DBOutput;

import com.loafle.overflow.module.apikey.dao.ApiKeyDAO;

import com.loafle.overflow.module.email.dao.EmailAuthDAO;
import com.loafle.overflow.module.member.dao.MemberDAO;

import com.loafle.overflow.module.noauthprobe.dao.NoAuthProbeDAO;
import com.loafle.overflow.module.probe.dao.ProbeDAO;
import com.loafle.overflow.module.sensor.dao.SensorDao;
import com.loafle.overflow.module.sensor.dao.SensorItemDao;
import com.loafle.overflow.module.target.dao.TargetDAO;
import io.grpc.ServerBuilder;
import org.codehaus.jackson.map.ObjectMapper;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.data.jpa.repository.JpaRepository;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

/**
 * Created by insanity on 17. 5. 23.
 */

public class DBProxy {

    private static final Logger logger = Logger.getLogger(DBProxy.class.getName());
    private io.grpc.Server server;

    private static Map<String, JpaRepository> daoMap = null;

    public DBProxy() {
        daoMap =  new ConcurrentHashMap();

        ApplicationContext ctx = new AnnotationConfigApplicationContext("com.loafle.overflow");

        daoMap.put("member", ctx.getBean(MemberDAO.class));
        daoMap.put("probe", ctx.getBean(ProbeDAO.class));
        daoMap.put("noauthAgent", ctx.getBean(NoAuthProbeDAO.class));
        daoMap.put("emailAuth", ctx.getBean(EmailAuthDAO.class));
        daoMap.put("sensor", ctx.getBean(SensorDao.class));
        daoMap.put("sensorItem", ctx.getBean(SensorItemDao.class));

//        daoMap.put("target", targetDAO);
//        daoMap.put("crawler", crawlerDAO);
//        daoMap.put("crawlerInputItem", crawlerInputItemDAO);
//        daoMap.put("crawlerInputItemMapping", crawlerInputItemMappingDAO);
//        daoMap.put("sensorItemCategory", sensorItemCategoryDao);
//        daoMap.put("sensorItemMapping", sensorItemMappingDao);
    }

    public void start(int port) throws IOException {
        server = ServerBuilder.forPort(port)
            .addService(new ServiceImpl())
            .build()
            .start();
        logger.info("Server started, listening on " + port);
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                // Use stderr here since the logger may have been reset by its JVM shutdown hook.
                System.err.println("*** shutting down gRPC server since JVM is shutting down");
                DBProxy.this.stop();
                System.err.println("*** server shut down");
            }
        });
    }

    public void stop() {
        if (server != null) {
            server.shutdown();
        }
    }

    static class ServiceImpl extends DBGrpc.DBImplBase {
        @Override
        public void exec(DBInput request,
                         io.grpc.stub.StreamObserver<com.loafle.overflow.db.api.DBOutput> responseObserver) {
            String targetDAO = request.getTargetDao();

            JpaRepository dao = daoMap.get(targetDAO);

            if(dao != null) {
                try {

                    String jsonResult = doQuery(request, dao);

                    DBOutput reply = DBOutput.newBuilder()
                        .setResult(jsonResult)
                        .build();
                    responseObserver.onNext(reply);
                    responseObserver.onCompleted();
                }catch(Exception e) {
                    e.printStackTrace();
                    responseObserver.onError(e);
                }

            }else {
                responseObserver.onError(new Exception("Not assigned DAO :" + targetDAO));
            }
        }

        private String doQuery(DBInput request, JpaRepository dao) throws Exception {

            String methodName = request.getMethod();
            Map<String, String> params = request.getParamsMap();

            ObjectMapper mapper = new ObjectMapper();

            Object retObj = null;
            Method method = null;

            if (methodName.equals("find")) {
                method = dao.getClass().getSuperclass().getMethod(methodName, String.class);
                retObj = method.invoke(dao, params.get("id"));
                return mapper.writeValueAsString(retObj);
            }

            List<Class> paramTypes = new ArrayList<Class>();
            List<Object> valueList = new ArrayList<Object>();
            for( String className : params.keySet() ){
                //className - model name(with package) , value - model JsonString
                Class<?> cls = Class.forName(className);
                Object obj = mapper.readValue((String)params.get(className), cls);
                valueList.add(obj);
                paramTypes.add(cls);
            }

            try {
                //Check if it belongs to the JPABaseDAO.
                method = dao.getClass().getSuperclass().getMethod(methodName, Object.class);
                retObj = method.invoke(dao, valueList.toArray(new Object[valueList.size()]));

            }catch(NoSuchMethodException e) {
                if(params.size() > 0) {
                    method = dao.getClass().getMethod(methodName, paramTypes.toArray(new Class[paramTypes.size()]));
                    retObj = method.invoke(dao, valueList.toArray(new Object[valueList.size()]));
                }else {
                    method = dao.getClass().getMethod(methodName);
                    retObj = method.invoke(dao);
                }
            }catch(InvocationTargetException e) {
                e.printStackTrace();
            }

            return mapper.writeValueAsString(retObj);
        }
    }

    public void blockUntilShutdown() throws InterruptedException {
        if (server != null) {
            server.awaitTermination();
        }
    }
}